# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 28 August 2022

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import copy as cp

from .fit_methods import fit_methods

class fit(fit_methods):
    def __init__(self,fit_module):

        # Tell the fit module that we are adding a gaussian fit
        fit_module.info['fit_dict']['Gaussian'] = self

        # Tell the fit module what parameters we want
        parameter_list = ['Standard Dev','A','Mu']

        # Set the model to be used
        self.model = self.ngaussian_model

        # Should there be a new axis added?
        self.new_ax = True

        # Setup the fit
        self.setup_fit(fit_module,parameter_list)

        return

    ##########################################
    ## THE MODEL
    ##########################################

    def gaussian(self,xx,sigma,a,mu):
        # A gaussian function used for curve fitting
        exp = -(1/2) * ((xx-mu)/sigma) ** 2
        return np.array(a * np.exp(exp))

    # Define an n gaussian model
    def ngaussian_model(self,xx,*params):
        y = np.zeros_like(xx)
        for i in range(0,len(params),3):
            y = np.add(y, self.gaussian(xx,params[i],params[i+1],params[i+2]), casting="unsafe")
        # return(y + yoff + slope * (xx-xoff))
        return y

    ##########################################
    ## PARAMETERS
    ##########################################

    # Method to determine guess parameters and bounds
    def get_params(self,parameters):
        #Set initial parameters and bounds

        p0_temp,ubound_temp,lbound_temp = [],[],[]

        for param in parameters:
            prms = parameters[param]
            p0_temp.append(prms['Estimate'])
            lbound_temp.append(prms['Lower Bound'])
            ubound_temp.append(prms['Upper Bound'])

        p0,ubound,lbound = [],[],[]

        # Combine the separate parameter arrays
        for i in range(self.fit_order):
            p0 = np.concatenate((p0,p0_temp))
            lbound = np.concatenate((lbound,lbound_temp))
            ubound = np.concatenate((ubound,ubound_temp))
        bounds = np.concatenate(([lbound],[ubound]),axis=0)

        # Return the initial parameter guesses and bounds
        return(p0,bounds)

    ##########################################
    ## PLOTTING METHODS
    ##########################################

    def plot_fit(self,fit_mod,stats,fit_number,ax,fit_ax=None):

        ##########################################
        ## PLOT THE FIT LINE ON THE OG AXIS
        ##########################################

        # Get the stats dictionary
        stats = fit_mod.data[ax]['Fit'][f'Fit {fit_number}']['Stats']
        xdata = stats['xdata']
        popt = stats['popt']

        # Define an array of hex colors
        colors = sns.color_palette(fit_mod.style['color_palette']).as_hex()

        # Plot the fit line
        fit_line = ax.plot(xdata,stats['simdat'],lw=2,alpha=0.6,color='red')

        ##########################################
        ## PLOT THE FIT LINE ON THE NEW AXIS
        ##########################################

        if fit_ax is not None:
            # Create a total area variable
            total_area = 0
            # Loop through each fit gaussian and save and plot it
            for ii in np.arange(self.fit_order):
                self.test = 5692
                # Add the data for each gaussian
                ydata_g = self.gaussian(xdata, popt[3*ii], popt[3*ii + 1], popt[3*ii + 2])
                fit_mod.data[fit_ax][f'Fit {ii+1}'] = {'xdata':xdata,'ydata':ydata_g}
                # Calculate the area
                self.test = 56922352
                area = stats['area'][f'fit{ii+1}']
                # Add the area to the total area
                total_area += area
                self.test = 56922341
                err = stats['area_err'][f'fit{ii+1}_err']
                fit_mod.artists[fit_ax][f'Fit {ii+1}'] = fit_ax.fill_between(
                            fit_mod.data[fit_ax][f'Fit {ii+1}']['xdata'],
                            fit_mod.data[fit_ax][f'Fit {ii+1}']['ydata'],
                            label=f'Fit {ii+1} - Area: {round(area,3)}, Error: {round(err,3)}',
                            color=colors[ii%6],lw = 1,alpha=0.6)
                self.test = 5692234215

            self.test = 523523
            # Add a legend
            fit_ax.legend()
            # Set the legend on the original axis
            fit_line[0].set_label(f'Area: {round(total_area,3)}')
            ax.legend()

        return fit_line

    ##########################################
    ## STATS
    ##########################################

    # Method to return stats about a certain fit
    def add_fit_stats(self,stats):

        self.test = 903874
        area,area_err,fwhm,fwhm_err = {},{},{},{}
        popt,perr = stats['popt'],stats['perr']

        self.test = 903874234
        # Calculate the area of a Gaussian (for example, if you want to calculate the column density)
        for i in range(0,self.fit_order):
            # area of the first Gaussian: area = sqrt(2pi)*width*amp
            area[f'fit{i+1}'] = np.sqrt(2*np.pi)*popt[i*3]*popt[i*3 + 1]
            self.test = 903874123
            # error of the area of the first Gaussian calculated from the errors in the Gaussian parameters
            area_err[f'fit{i+1}_err'] = area[f'fit{i+1}'] * np.sqrt((perr[i*3]/popt[i*3])**2 + (perr[i*3 + 1]/popt[i*3 + 1])**2)

            self.test = stats
            # The FWHM (full width half maximum) of the Gaussian can be calculated from the width parameter of the Gaussian
            fwhm[f'fit{i+1}'] = 2.35482 * popt[3*i]
            fwhm_err[f'fit{i+1}_err'] = 2.35482 * perr[3*i]

        self.test = 90387423
        stats['area'],stats['area_err'],stats['fwhm'],stats['fwhm_err'] = area,area_err,fwhm,fwhm_err

        return(stats)



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 26 September 2022

import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy

from scipy.optimize import curve_fit
import seaborn as sns

class fit_methods():

    def __init__(self):

        # List of parameter strings
        parameter_list = []

        # A dictionary is created to hold the parameters
        self.parameters = self.setup_parameters(parameter_list)

        # The model function below is saved as the model
        self.model = self.defined_model

        # New axis
        self.new_ax = False

        return

    ##########################################
    ## SETUP THE CLASS
    ##########################################

    def setup_fit(self,fit_module,parameter_list):

        # Create a fit parameters dictionary from the parameter list
        fit_module.info['fit_params'] = dict.fromkeys(parameter_list,{})
        for key in fit_module.info['fit_params']:
            fit_module.info['fit_params'][key] = {'Estimate':5,'Lower Bound':0,'Upper Bound':10}

        # Get the fit order from the info dictionary
        self.fit_order = fit_module.info['fit_order']

        # Calculate the parameters
        self.parameters = self.get_params(fit_module.info['fit_params'])

    ##########################################
    ## IMPORTED FITS
    ##########################################

    def fits(self): return

    ##########################################
    ## PARAMETERS
    ##########################################

    # Method to determine guess parameters and bounds
    def get_params(self,parameters):
        #Set initial parameters and bounds

        p0,ubound,lbound = [],[],[]

        for param in parameters:
            prms = parameters[param]
            p0.append(prms['Estimate'])
            lbound.append(prms['Lower Bound'])
            ubound.append(prms['Upper Bound'])

        return p0,[lbound,ubound]

    ##########################################
    ## THE MODEL
    ##########################################

    def defined_model(self,xx,*parameters): return xx

    ##########################################
    ## FITTING METHOD
    ##########################################

    def fit(self,fit_struct,ax,fit_info,new_ax=False):

        ##########################################
        ## PULL DATA
        ##########################################

        # Determine the primary curve for the current axis
        curve = self.artists_global['Primary Artists'][ax]
        # Separate the xdata values
        xdata = self.artists_global[ax][curve].get_data()[0]

        ##########################################
        ## DETERMINE WHICH FIT IS BEING PLOTTED
        ##########################################

        # Initialize a line number
        fit_number = 1
        # Determine which fit this is
        for key in self.artists[ax].keys():
            if 'Fit Line' in key: fit_number += 1

        # If there should be a new axis added
        if new_ax:
            ##########################################
            ## SETUP AXES
            ##########################################

            # Add an axis to the figure
            self.resize_figure(len(self.fig.axes)+1,self.fig)
            # Save that axis
            fit_ax = self.fig.axes[-1]
            self.data[ax]['Fit']['Axis'].append(fit_ax)

            ##########################################
            ## SETUP DICTIONARIES
            ##########################################

            # Add a plot on which to display the fit
            self.artists[fit_ax] = {}
            # Set the primary figure artist to the Fit Line
            self.artists_global['Primary Artists'][fit_ax] = 'Fit Line'
            # Create a data dictionary for the new axis
            self.data[fit_ax] = {
            'FitRange': self.data[ax]['FitRange'],
            'Data Axis':ax,
            'Fit Number': fit_number
            }
        else: fit_ax = None

        ##########################################
        ## FIT MODEL
        ##########################################

        # Fit n gaussians to the primary data on the selected plot
        stats = self.calculate_fit(ax,fit_struct.model,fit_struct.parameters)

        # Add optional extra statistics
        stats = fit_struct.add_fit_stats(stats)

        self.test = stats
        self.data[ax]['Fit'][f'Fit {fit_number}'] = {'Stats':stats}

        ##########################################
        ## PLOT THE MODEL
        ##########################################

        # Save the Fit Line data to the original axis
        self.data[ax]['Fit'][f'Fit {fit_number}']['Fit Line'] = {'xdata':xdata,'ydata':stats['simdat']}

        # Save the fit type and range to the original axis
        self.data[ax]['Fit'][f'Fit {fit_number}']['fit_type'] = fit_info['fit_type']
        self.data[ax]['Fit'][f'Fit {fit_number}']['fit_params'] = deepcopy(fit_info['fit_params'])
        self.data[ax]['Fit'][f'Fit {fit_number}']['fit_order'] = fit_info['fit_order']
        self.data[ax]['Fit'][f'Fit {fit_number}']['FitRange'] = self.data[ax]['FitRange']

        # Plot the optimal parameters
        self.artists[ax][f'Fit Line {fit_number}'] = fit_struct.plot_fit(self,stats,fit_number,ax,fit_ax)[0]


    ##########################################
    ## THE MODEL FIT
    ##########################################

    def calculate_fit(self,ax,model,parameters):

        ##########################################
        ## DEFINE DATA
        ##########################################

        # Determine the primary curve for the current axis
        curve = self.artists_global['Primary Artists'][ax]
        # Pull out the x and y data from that curve
        xdata,ydata = self.artists_global[ax][curve].get_data()
        # Determine the range of the fit
        xmin,xmax = sorted(self.data[ax]['FitRange'])

        p0,bounds = parameters

        ##########################################
        ## SUBSET THE DATA TO THE INPUT RANGE
        ##########################################

        # Determine the indices of the data values within the range
        ii = np.squeeze(np.where((xdata > xmin) & (xdata < xmax)))
        # Define subsets of the data arrays
        xsub,ysub = xdata[ii],ydata[ii]

        ##########################################
        ## FIT THE MODEL
        ##########################################

        # Fit a curve
        popt,pcov = curve_fit(model, xsub, ysub, p0=p0, bounds=bounds)

        ##########################################
        ## CALCULATE THE FIT STATS
        ##########################################

        # pcov is a matrix with values related to the errors of the fit.
        # To get the actual errors of the Gaussian parameters one needs to calculate the square root of the values in the diagonal

        # Create a fit from the xdata and fitted parameters
        simdat = model(xdata,*popt)
        # Calculate the errors of the fit
        perr = np.sqrt(np.diag(pcov))
        # Save the relevant statistics
        stats = {
        'popt':popt,'perr':perr,
        'xsub':xsub,'ysub':ysub,
        'xdata':xdata,
        'simdat':simdat
        }

        # Return the stats dict
        return(stats)

    ##########################################
    ## PLOT THE FIT
    ##########################################

    def plot_fit(self,fit_mod,stats,fit_number,ax,fit_ax=None):

        # Get the stats dictionary
        stats = fit_mod.data[ax]['Fit'][f'Fit {fit_number}']['Stats']
        xdata = stats['xdata']

        # Plot the fit line
        fit_line = ax.plot(xdata,stats['simdat'],lw=2,alpha=0.6,color='red')

        return fit_line

    ##########################################
    ## STATS
    ##########################################

    # Method to return stats about a certain fit
    def add_fit_stats(self,stats): return stats



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 28 August 2022

class fit_gaussian(fit):
    def __init__(self,fig,menu=None,data=None,artists_global=None,data_global=None,load=None):


    ##########################################
    ## THE MODEL
    ##########################################

    def gaussian(self,xx,sigma,a,mu):
        # A gaussian function used for curve fitting
        exp = -(1/2) * ((xx-mu)/sigma) ** 2
        return np.array(a * np.exp(exp))

    # Define an n gaussian model
    def ngaussian_model(self,xx,*params):
        y = np.zeros_like(xx)
        for i in range(0,len(params),3):
            y = np.add(y, self.gaussian(xx,params[i],params[i+1],params[i+2]), casting="unsafe")
        # return(y + yoff + slope * (xx-xoff))
        return y

    ##########################################
    ## PARAMETERS
    ##########################################

    # Method to determine guess parameters and bounds
    def get_params(self,xmin,xmax,ymax,gnum=1):
        #Set initial parameters and bounds

        xavg = (xmin+xmax)/2
        std_max = self.info['std_max']

        p0,bound1,bound2 = [],[],[]

        #             std        a        mu
        p0_fit   =  [ std_max/2, ymax,  xavg ]   # GUESSES
        bnd_fit1 =  [ 0,         1e-6,    xmin ]  # LOWER BOUNDS
        bnd_fit2 =  [ std_max,   ymax*2, xmax ]  # UPPER BOUNDS

        # Combine the separate parameter arrays
        for i in range(gnum):
            p0 = np.concatenate((p0,p0_fit))
            bound1 = np.concatenate((bound1,bnd_fit1))
            bound2 = np.concatenate((bound2,bnd_fit2))
        bounds = np.concatenate(([bound1],[bound2]),axis=0)

        # Return the initial parameter guesses and bounds
        return(p0,bounds)

    ##########################################
    ## STATS
    ##########################################

    # Method to return stats about a certain fit
    def add_fit_stats(self,stats,gnum):

        area,area_err,fwhm,fwhm_err = {},{},{},{}
        popt,perr = stats['popt'],stats['perr']

        # Calculate the area of a Gaussian (for example, if you want to calculate the column density)
        for i in range(0,gnum):
            # area of the first Gaussian: area = sqrt(2pi)*width*amp
            area[f'fit{i+1}'] = np.sqrt(2*np.pi)*popt[i*3]*popt[i*3 + 1]
            # error of the area of the first Gaussian calculated from the errors in the Gaussian parameters
            area_err[f'fit{i+1}_err'] = area[f'fit{i+1}'] * np.sqrt((perr[i*3]/popt[i*3])**2 + (perr[i*3 + 1]/popt[i*3 + 1])**2)

            # The FWHM (full width half maximum) of the Gaussian can be calculated from the width parameter of the Gaussian
            fwhm[f'fit{i+1}'] = 2.35482 * popt[3*i]
            fwhm_err[f'fit{i+1}_err'] = 2.35482 * perr[3*i]

        stats['area'],stats['area_err'],stats['fwhm'],stats['fwhm_err'] = area,area_err,fwhm,fwhm_err

        return(stats)



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 22 September 2022

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import ipywidgets as widgets
import functools
import pickle as pkl

import pandas as pd

from .base import widget_base

class export(widget_base):
    def __init__(self,fig,main_menu,menu=None,param_menu=None,exp_path=None):
        '''
        '''
        self.widget_on = False
        self.main_menu = main_menu
        self.fig = fig

        # Initialize defaults
        self.info = self.setup_defaults(exp_path)

        # Initialize all buttons
        self.button_list,self.toggle_buttons,self.param_buttons = self.setup_buttons()
        # Place all the menu buttons
        if menu is None:
            self.menu = widgets.HBox()
            self.place_menu(self.menu,self.button_list)
        else: self.menu = menu
        if param_menu is None:
            self.param_menu = widgets.HBox()
            self.place_menu(self.param_menu,self.param_buttons)
        else: self.param_menu = param_menu

    def setup_defaults(self,exp_path):

        if exp_path is None: exp_path = ''
        info = {
        'exp_path':exp_path,
        'dpi':200,
        'axis': self.fig.axes[0],
        'param_show':False,
        'datadict_path':[]
        }

        return info

    def setup_buttons(self):

        ##########################################
        ## INTERACTIVE PLOTTING BUTTONS
        ##########################################

        # Add a button widget to add new points
        self.export_path_text = widgets.Text(description='Path:',value=self.info['exp_path'],style={'description_width': 'initial'},layout=widgets.Layout(width='200px'))

        def export_path_entered(b, self = self):

            self.info['exp_path'] = self.export_path_text.value

        self.export_path_text.observe(functools.partial(export_path_entered, self=self))

        #####################

        # Add a button widget to add new points
        self.export_all_button = widgets.Button(description='Export Figure')

        def on_export_all_button_clicked(b, self = self):

            exp_dict = self.main_menu.export_dict()
            pkl.dump(exp_dict,open(f"{self.info['exp_path']}.pkl",'wb'))

        self.export_all_button.on_click(functools.partial(on_export_all_button_clicked, self=self))

        #####################

        # Add a button widget to save specfic data
        self.export_data_button = widgets.ToggleButton(description='Export Data')

        def on_export_data_button_clicked(b, self = self):

            if self.info['param_show']:
                self.info['param_show'] = False
                self.place_menu(self.param_menu,[])
            else:
                self.info['param_show'] = True
                self.param_buttons[0].options = ['All'] + list(self.main_menu.get_data().keys())
                self.place_menu(self.param_menu,[self.param_buttons[0]]+[self.save_data_button])

        self.export_data_button.observe(functools.partial(on_export_data_button_clicked, self=self))

        #####################

        # Add a button widget to save the figure
        self.save_button = widgets.Button(description='Save Plot')

        def on_save_button_clicked(b, self = self):

            plt.savefig(f"{self.info['exp_path']}.png",dpi=self.info['dpi'])

        self.save_button.on_click(functools.partial(on_save_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.dpi_text = widgets.IntText(description='DPI:',value=self.info['dpi'],style={'description_width': 'initial'},layout=widgets.Layout(width='100px'))

        def dpi_entered(b, self = self):

            self.info['dpi'] = self.dpi_text.value

        self.dpi_text.observe(functools.partial(dpi_entered, self=self))

        #####################

        self.param_selects = []



        #####################

        # Add a button widget to save the figure
        self.save_data_button = widgets.Button(description='Save Data')

        def on_save_data_button_clicked(b, self = self):

            export_dict = self.main_menu.get_data(*self.info['datadict_path'])
            if type(export_dict) == dict:
                export_df = self.df(export_dict)
            elif type(export_dict) == list or type(export_dict) == np.ndarray:
                export_df = pd.DataFrame()
                export_df[self.info['datadict_path'][-1]] = export_dict
            else: return

            path = self.info['exp_path']

            # If there is no path, do nothing
            if path == '': return
            # If there is no specified extension, default to csv
            if '.' not in path: path += '.csv'

            # Save the data to the desired filetype
            if path.split('.')[-1] == 'csv':
                export_df.to_csv(path)
            elif path.split('.')[-1] == 'dpt':
                self.save_dpt(path,export_df)

        self.save_data_button.on_click(functools.partial(on_save_data_button_clicked, self=self))

        #####################

        return [self.export_path_text,self.export_data_button,self.export_all_button,self.save_button,self.dpi_text],[self.export_data_button],self.param_selects

    def update(self,artists,data): return

    # Define a function used to setup the export buttons
    def data_select(self,param_selects,index):
        # Define a function for the first few buttons
        def func(b,self=self):
            # Make sure the function does not repeatedly get called
            if param_selects[index].value is None: return
            elif param_selects[index].value == 'All':
                self.place_menu(self.param_menu,param_selects[:index+1]+[self.save_data_button])
                return
            # Set the param_select values of the deeper selects to None
            for select in param_selects[index+1:]: select.value = None

            # Clear the data path following the current entry
            self.info['datadict_path'] = self.info['datadict_path'][:index]
            # Add the current entry to the data path
            self.info['datadict_path'].append(param_selects[index].value)

            # Define a subset of the data
            sub_data = self.main_menu.get_data(*self.info['datadict_path'])
            self.sub_data = sub_data
            self.index = index
            self.param_selects = param_selects
            if type(sub_data) == dict and index < len(param_selects)-1:
                # For the next selection, set the option to the keys of the selected dictionary
                param_selects[index+1].options = ['All'] + list(sub_data.keys())
                # Place an updated menu
                self.place_menu(self.param_menu,param_selects[:index+2]+[self.save_data_button])
            else:
                # Place an updated menu
                self.place_menu(self.param_menu,param_selects[:index+1]+[self.save_data_button])
                pass

        return func

    # Override activation method
    def activate(self):
        # Activate the widget
        self.widget_on = True
        self.place_menu(self.menu,self.button_list)
        # Setup parameter selects for saving data
        self.param_selects = []
        for num in range(self.dict_depth(self.main_menu.get_data())):
            button = widgets.Select(options=[],rows=0,style={'description_width': 'initial'},layout = widgets.Layout(width='150px'))
            self.param_selects.append(button)
        # Update the param_buttons as well
        self.param_buttons = self.param_selects
        for index,select in enumerate(self.param_selects):
            select.observe(self.data_select(self.param_selects,index))
        # Return the widget toggle
        return self.widget_on



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 22 August 2022

import numpy as np
import matplotlib.pyplot as plt
import ipywidgets as widgets

import pandas as pd

class widget_base:
    '''
    This class serves as the base class for all of the astrochem-tools
    interactive classes. The methods outlined in this class will be inhereted
    and modified to enable different interactive elements in the other widgets.
    '''
    def __init__(self,fig=None,ax=None): return

    def __call__(self,event): return

    ##########################################
    ## SETUP METHODS
    ##########################################

    def setup_buttons(self): return

    # Update the buttons onscreen
    def place_menu(self,menu,button_list):
        # Place the buttons and menus into a box and display it
        menu.children = button_list
        # Set the output
        output = widgets.Output()
        # Display the box
        display(menu, output)
        self.fig.canvas.draw_idle()

    # Initialize Artists
    def define_artists(self): return

    ##########################################
    ## MATH METHODS
    ##########################################

    # Calculate the eucldian distance of a point to a series of points
    def bigeuc(self,x0,y0,xarr,yarr,ax):
        # Convert array variables into numpy arrays for arithmetic
        xarr = np.array(xarr); yarr = np.array(yarr)
        # Find the boundaries of the axis
        xmin,xmax = ax.get_xlim(); ymin,ymax = ax.get_ylim()
        # Subtract off the minima to set the lower bound at 0
        x0 -= xmin; xarr -= xmin; xmax -= xmin
        y0 -= ymin;
        yarr -= ymin;
        ymax -= ymin
        # Normalize all values to range from 0 to 1
        x0 /= xmax; xarr /= xmax; y0 /= ymax; yarr /= ymax
        # Return the euclidian distances of the normalized points
        return [self.euclid(x0,y0,xarr[ii],yarr[ii]) for ii in range(len(xarr))]

    # Calculate the euclidian distance between two points
    def euclid(self,x0, y0, x1, y1):

        return np.sqrt((x1 - x0)**2 + (y1-y0)**2)

    # Calculate the horizontal distance between a point and an array of points
    def hdist(self,x0,xarr,ax):
        return self.bigeuc(x0,0,xarr,np.zeros_like(xarr),ax)

    # Find a y value for an x value along a line defined by two points
    def lincalc(self,x1,y1,x2,y2,x3):
        m = (y2-y1)/(x2-x1)
        y3 = m * (x3-x1) + y1
        return y3

    ##########################################
    ## SUPPORT METHODS
    ##########################################

    # Calculate ordered arrays from the xdat and ydat arrays
    def order(self,xdata,ydata):
        # Sort the data into tuples so that the line connects linearly
        xytuples = sorted([i for i in zip(xdata,ydata)])
        # Create new arrays to store the separated x and y values
        xord,yord = [],[]
        # Separate the ordered x and y values
        for x,y in xytuples: xord.append(x); yord.append(y)

        return(xord,yord)

    # Listen for a button press event
    def listen(self,widget):
        # Call a listener on the point
        self.cid = self.fig.canvas.mpl_connect('button_press_event',widget)

    # Clear a plot
    def clear(self,line,vlines=False,ax=None):
        # If there was a newpoint added
        if line != None and not vlines:
            # Remove all points
            line.set_data([],[])
        if vlines:
            # Remove all vlines
            self.set_segments(line,[],ax)

    ##########################################
    ## FIGURE METHODS
    ##########################################

    def fig_init(self,fig,ax):
        if (fig is None) or (ax is None):
            # Close any past figures
            plt.close()
            # Define the default figure size for the plot
            self.figsize = [10,3.5]
            # Define new axes
            self.fig, self.ax = plt.subplots(figsize=self.figsize)
            self.fig.tight_layout()
        else:
            # Use predefined fig and ax
            self.fig,self.ax = fig,ax
            # Determine the user-defined figure size
            self.figsize = self.fig.get_size_inches()

    def resize_figure(self,nplots,fig):
        '''
        Changes the number of vertical axes in a matplotlib figure
        with dimensions 10 by 3.5 * nplots inches and adjusts the
        figure size and axes placement automatically

        Parameters
        ----------
        nplots : int
            An integer denoting the final number of axes the figure
            will have after resizing
        fig : matplotlib.figure
            The matplotlib figure object being resized
        axes : matplotlib.axes or [matplotlib.axes, ...]
            The matplotlib axes object or list of matplotlib axes objest
            being resized
        '''
        # Create axes variable
        axes = fig.axes
        # Create a temporary figure
        temp_fig,temp_axes = plt.subplots(nplots,1,figsize=(10,3.5*nplots))
        temp_fig.tight_layout()
        # Save bounds from that temporary figure
        if nplots==1: bounds = [temp_axes.get_position().bounds]
        else: bounds = [temp_ax.get_position().bounds for temp_ax in temp_axes]
        # If there is only one axis
        if str(type(axes)) == "<class 'matplotlib.axes._subplots.AxesSubplot'>":
            # Resize it
            axes.set_position(bounds[0])
            # Resize the rest of the axes
            for bound in bounds[1:]: fig.add_axes(bound)
        # If there are more than one axis
        else:
            # If the number of axes are being reduced
            if len(axes) > nplots:
                # Delete all the extra axes
                for ax in axes[nplots:]: fig.delaxes(ax)
                # Redefine the input to a truncated version
                axes = fig.axes
            # Resize the rest of the axes
            for ind,ax in enumerate(axes): ax.set_position(bounds[ind])
            for bound in bounds[len(axes):]: fig.add_axes(bound)
        # Resize the figure
        fig.set_size_inches(10,3.5*nplots)
        # Close the temporary figure
        plt.close(temp_fig)
        # Return the new axes
        return([fig.get_axes(),bounds,axes,nplots])

    def set_segments(self,vline,arr,ax):
        ymin,ymax = ax.get_ylim()
        newsegs = [[[xval,ymin],[xval,ymax]] for xval in arr]
        vline.set_segments(newsegs)

    def update(self,artists,data):
        '''
        Called every time a new menu is selected
        It should pull the artists and data into the primary dicts
        The artist_ and data_ push methods should be called as each are being
        updated, so the primaries should always be kept up to date
        Just remember to delete any artists/data from both local and primary dicts
        '''
        # Update the global artist and data dictionaries
        self.push_artists(artists,self.artists)
        self.push_data(data,self.data)

    def push_artists(self,artists_global,artists):
        for ax in artists:
            if ax not in artists_global.keys(): artists_global[ax] = {}
            for artist in artists[ax]:
                if isinstance(ax,str): artists_global[ax] = artists[ax]
                else: artists_global[ax][artist] = artists[ax][artist]

    def pull_artists(self,artists=None):
        if artists is None: return {'Primary Artists':{},'Interactive Axes':{}}
        else: return artists

    def push_data(self,data_global,data):
        for ax in data:
            if ax not in data_global.keys(): data_global[ax] = {}
            for dat in data[ax]:
                data_global[ax][dat] = data[ax][dat]

    def pull_data(self,data=None):
        if data is None: return {'Primary Artists':{},'Interactive Axes':{}}
        else: return data

    def update_dictionaries(self,ax,artist_key,data_key):
        # LOCAL ARTISTS
        if ax not in self.artists.keys():
            # Add the axis to the artist dict
            self.artists[ax] = self.new_artists(ax,self.style)
        elif artist_key not in self.artists[ax].keys():
            # Else, merge the dictionary with a new one
            self.artists[ax] = self.artists[ax] | self.new_artists(ax,self.style)
        # GLOBAL ARTISTS
        if ax not in self.artists_global.keys():
            # Add the axis to the artist dict
            self.artists_global[ax] = self.new_artists(ax,self.style)
        elif artist_key not in self.artists_global[ax].keys():
            # Else, merge the dictionary with a new one
            self.artists_global[ax] = self.artists_global[ax] | self.new_artists(ax,self.style)
        # LOCAL DATA
        if ax not in self.data.keys():
            # Add the axis to the data dict
            self.data[ax] = self.data_init()
            self.data_global[ax] = self.data_init()
        elif data_key not in self.data[ax].keys():
            # Else, merge the dictionary with a new one
            self.data[ax] = self.data[ax] | self.data_init()
        # GLOBAL DATA
        if ax not in self.data_global.keys():
            # Add the axis to the data dict
            self.data_global[ax] = self.data_init()
        elif data_key not in self.data_global[ax].keys():
            # Else, merge the dictionary with a new one
            self.data_global[ax] = self.data_global[ax] | self.data_init()

    def printdict(self,dictionary,indent=''):
        if indent=='': print('Dictionary')
        # If the input is a dictionary
        if isinstance(dictionary,dict):
            # Save the list of keys
            keys = list(dictionary.keys())
            # Set the indent strings
            tab = ' |  '
            turn = ' |--'
            # For each key of the dictionary
            for key in keys:
                # # Check if it is the final key
                if key == keys[-1]:
                    # Set special indent strings
                    tab = '    '
                    turn = " '--"
                # Print the dictionary key
                print(f'{indent}{turn}{key}')
                # Check to see if it has any subdictionaries
                self.printdict(dictionary[key],indent+tab)
        # Once all sub-dictionaries have been printed, return
        else: return

    def fig_show(self,fig):
        # Create a new figure
        figure = plt.figure()
        # Create a new manager
        manager = figure.canvas.manager
        # Overwrite the figure in the manager
        manager.canvas.figure = fig
        # Set the figure canvas to the manager's canvas
        fig.set_canvas(manager.canvas)

# ADMIN METHODS

    def activate(self):
        # Activate the widget
        self.widget_on = True
        self.place_menu(self.menu,self.button_list)
        # Return the widget toggle
        return self.widget_on

    def deactivate(self,main=False):
        # Deactivate the widget
        self.widget_on = False
        # If not the main menu
        if not main:
            # Clear the widget menu
            self.place_menu(self.menu,[])
        # Return the widget toggle
        return self.widget_on

# FUNCTIONAL METHODS

    def dict_depth(self,dct):
        depths = []
        for key in dct:
            if type(dct[key]) == dict:
                depths.append(self.dict_depth(dct[key]))
        if len(depths) == 0: depths.append(0)
        return max(depths)+1

# PRINT AND RETURN METHODS

    def data_tree(self,*kwargs):
        ddict = self.get_data()
        for arg in kwargs:
            if arg in ddict.keys():
                ddict = ddict[arg]
        self.printdict(ddict)

    def artist_tree(self,*kwargs):
        adict = self.get_artists()
        for arg in kwargs:
            if arg in adict.keys():
                adict = adict[arg]
        self.printdict(adict)

    def get_data(self,*kwargs):
        # Get the list of axes
        axes = self.get_axes()
        # Create a new data dictionary for data export
        exp_data = {}
        # Fill the new dictionary
        for ind,ax in enumerate(axes):
            if ax in self.data:
                exp_data[f'Axis {ind+1}'] = self.data[ax]
        for arg in kwargs:
            if arg in exp_data.keys():
                exp_data = exp_data[arg]
        # Return the export dictionary
        return exp_data

    def get_artists(self,*kwargs):
        # Get the list of axes
        axes = self.get_axes()
        # Create a new data dictionary for data export
        exp_artists = {}
        # Fill the new dictionary
        for ind,ax in enumerate(axes):
            exp_artists[f'Axis {ind+1}'] = self.artists[ax]
        for arg in kwargs:
            if arg in exp_artists.keys():
                exp_artists = exp_artists[arg]
        # Return the export dictionary
        return exp_artists

    def get_fig(self):
        return self.fig

    def get_axes(self):
        return self.fig.axes

    def export_dict(self):
        return {}

    def df(self,datadict,dataframe=pd.DataFrame(),path='',collist=[]):
        '''
        Create a dataframe from a Dictionary
        Recursive method
        '''
        for key in datadict:
            data = datadict[key]
            if type(data) == dict:
                # If the data is a dictionary, iterate recusively
                dataframe = self.df(datadict[key],dataframe,f'{path} - {key}',collist=collist)
            elif type(data) == list or type(data) == type(np.array([])):
                # Create a new dataframe with the list
                newdf = pd.DataFrame(data,columns=[f'{path} - {key}'[3:]])
                cols = list(dataframe.columns)
                # Merge the shorter dataframe to the larger one
                if len(dataframe) > len(newdf): dataframe = dataframe.join(newdf)
                else: dataframe = newdf.join(dataframe)
                # Update the list of columns
                cols.append(list(newdf.columns)[0])
                # Reorder the dataframe
                dataframe = dataframe[cols]
        # Return the final datafame
        return dataframe

    ##########################################
    ## FILETYPE EXPORTS
    ##########################################

    def save_dpt(self,filename,dataframe):
        # Check to make sure the data is of the right type
        if type(dataframe) == dict:
            dataframe = self.df(dataframe)
            self.save_dpt(filename,dataframe)
        elif type(dataframe) == list or type(dataframe) == np.ndarray:
            dataframe = pd.DataFrame()
            dataframe['data'] = dataframe
            self.save_dpt(filename,dataframe)
        elif type(dataframe) == type(pd.DataFrame()):
            with open(filename, "w") as newFile:
                for index in range(len(dataframe)):
                    dataline = ''
                    for point in dataframe.iloc[index]:
                        dataline += f'{point} '
                    newFile.write(f'{dataline[:-1]}\n')
        else: print('Please enter valid data')



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 28 August 2022

import numpy as np
import matplotlib.pyplot as plt
import ipywidgets as widgets
import functools

from scipy.optimize import curve_fit
import seaborn as sns
import copy as cp

from .base import widget_base
from .fit_methods import fit_methods

from .gaussian_fit import fit as gaussian_fit

class fit(widget_base,fit_methods):

    def fits(self):
        gaussian_fit(self)

    def __init__(self,fig,menu=None,param_menu=None,data=None,artists_global=None,data_global=None,load_data=None):
        '''
        '''
        # Initialize the widget in the off position
        self.widget_on = False
        # Set the display setting to display output
        self.show = True
        # Setup the figure
        self.fig = fig
        # Setup possible load data
        if load_data is not None:
            if 'fit' in load_data: load_data = load_data['fit']

        # Initialize all global matplotlib artists
        self.artists_global = self.pull_artists(artists_global)
        # Initialize all global data
        self.data_global = self.pull_data(data_global)
        # Create a local artist dictionary
        self.artists = {}

        # Initialize defaults
        self.style,self.info,self.data = self.setup_defaults(load_data)

        # Setup the fits
        self.fits()

        # Initialize all buttons
        self.button_list,self.toggle_buttons,self.param_button_list = self.setup_buttons()
        # Place all the menu buttons
        if menu is None:
            self.menu = widgets.HBox()
            self.place_menu(self.menu,self.button_list)
        else: self.menu = menu
        if param_menu is None:
            self.param_menu = widgets.HBox()
            self.place_menu(self.param_menu,self.param_button_list)
        else: self.param_menu = param_menu

        if load_data is not None: self.load(load_data)

    def setup_defaults(self,load_data=None):

        style = {
        'fill_color':'deepskyblue',
        'alpha':0.6,
        'vline_style':'--',
        'vline_width':2,
        'color_palette':'Set2'}

        info = {
        # Initialize the add, move, and delete
        'interactive_mode':'off',
        # Initialize an active axis
        'active_ax':None,
        # Initialize a selected boolean
        'selected':False,
        # Set an interactive click distance
        'click_dist':0.02,
        # Initialize a list of fit types
        'fit_dict':{},
        # Initialize a parameter:
        'parameter':'',
        # Initialize a fit type
        'fit_type':'None',
        # Initialize a parameters dict
        'fit_params':{},

        ##### TO BE REMOVED
        # Initialize a fit order
        'fit_order':1,
        # Initialize a fit std range
        'std_max':10
        }

        data = dict.fromkeys(self.artists_global['Interactive Axes'],self.data_init())

        if load_data is not None:
            load_data['info']['interactive_mode'] = 'off'
            load_data['info']['selected'] = False
            load_data['info']['active_ax'] = None

            return load_data['style'],load_data['info'],data

        return style,info,data

    def setup_parameters(self,params):
        param_dict = {}
        for param in params:
            param_dict[param] = {
            'Estimate':0,
            'Lower Bound':0,
            'Upper Bound':0
            }
        return param_dict

    def setup_buttons(self):

        ##########################################
        ## INTERACTIVE PLOTTING BUTTONS
        ##########################################

        # Add a button widget to add new points
        self.add_range_button = widgets.ToggleButton(description='Add Range')

        def on_add_range_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.add_range_button: button.value = False

            # Turn on add_range mode
            if self.info['interactive_mode'] == 'add_range':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'add_range'

        self.add_range_button.observe(functools.partial(on_add_range_button_clicked, self=self))

        #####################

        # Add a button widget to adjust current points
        self.adjust_range_button = widgets.ToggleButton(description='Adjust Range')

        def on_adjust_range_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.adjust_range_button: button.value = False

            # Turn on adjust_range mode
            if self.info['interactive_mode'] == 'adjust_range':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'adjust_range'

        self.adjust_range_button.observe(functools.partial(on_adjust_range_button_clicked, self=self))

        #####################

        # Add a selector widget to change the fit type
        self.fit_type_select = widgets.Select(description='Fit:',options=list(np.concatenate([['None'],list(self.info['fit_dict'].keys())])),rows=0,style={'description_width':'initial'},layout = widgets.Layout(width='200px'))

        def fit_type_select_clicked(b, self = self):

            self.info['fit_type'] = self.fit_type_select.value

            if self.fit_type_select.value == 'None':
                # Clear the parameter menu
                self.place_menu(self.param_menu,[])
                return

            if len(self.info['fit_params'].keys()) > 0:
                # Display the paramter menu
                self.place_menu(self.param_menu,self.param_button_list)
                self.info['parameter'] = self.parameter_select.value

        self.fit_type_select.observe(functools.partial(fit_type_select_clicked, self=self))

        #####################

        # Add an integer text widget to change the fit order
        self.fit_order_inttext = widgets.IntText(description='Order:',value=1,style={'description_width':'initial'},layout = widgets.Layout(width='95px'))

        def fit_order_inttext_clicked(b, self = self):

            # Save the fit order
            self.info['fit_order'] = self.fit_order_inttext.value
            # Retrieve the fit module
            fit_module = self.info['fit_dict'][self.info['fit_type']]
            # Set the fit order in the module
            fit_module.fit_order = self.info['fit_order']
            # Recalculate the parameters
            fit_module.parameters = fit_module.get_params(self.info['fit_params'])

        self.fit_order_inttext.observe(functools.partial(fit_order_inttext_clicked, self=self))

        #####################

        # Add a selector widget to change the input paramters
        self.parameter_select = widgets.Select(description='Paramters:',options=list(self.info['fit_params'].keys()),rows=0,style={'description_width':'initial'},layout = widgets.Layout(width='200px'))

        def parameter_select_clicked(b, self = self):

            self.info['parameter'] = self.parameter_select.value

            # Retrieve the current values of the fit parameters
            self.estimate_inttext.value = self.info['fit_params'][self.info['parameter']]['Estimate']
            self.lower_bound_inttext.value = self.info['fit_params'][self.info['parameter']]['Lower Bound']
            self.upper_bound_inttext.value = self.info['fit_params'][self.info['parameter']]['Upper Bound']

        self.parameter_select.observe(functools.partial(parameter_select_clicked, self=self))

        #####################

        # Add an integer text widget to set a paramter estimate
        self.estimate_inttext = widgets.IntText(description='Estimate:',value=5,style={'description_width':'initial'},layout = widgets.Layout(width='120px'))

        def estimate_inttext_clicked(b, self = self):

            # Save the fit parameter
            self.info['fit_params'][self.info['parameter']]['Estimate'] = cp.deepcopy(self.estimate_inttext.value)
            # Retrieve the fit module
            fit_module = self.info['fit_dict'][self.info['fit_type']]
            # Recalculate the parameters
            fit_module.parameters = fit_module.get_params(self.info['fit_params'])

        self.estimate_inttext.observe(functools.partial(estimate_inttext_clicked, self=self))

        #####################

        # Add an integer text widget to set the lower paramter bound
        self.lower_bound_inttext = widgets.IntText(description='Lower Bound:',value=0,style={'description_width':'initial'},layout = widgets.Layout(width='150px'))

        def lower_bound_inttext_clicked(b, self = self):

            # Save the fit parameter
            self.info['fit_params'][self.info['parameter']]['Lower Bound'] = cp.deepcopy(self.lower_bound_inttext.value)
            # Retrieve the fit module
            fit_module = self.info['fit_dict'][self.info['fit_type']]
            # Recalculate the parameters
            fit_module.parameters = fit_module.get_params(self.info['fit_params'])

        self.lower_bound_inttext.observe(functools.partial(lower_bound_inttext_clicked, self=self))

        #####################

        # Add an integer text widget to set the upper paramter bound
        self.upper_bound_inttext = widgets.IntText(description='Upper Bound:',value=10,style={'description_width':'initial'},layout = widgets.Layout(width='150px'))

        def upper_bound_inttext_clicked(b, self = self):

            # Save the fit parameter
            self.info['fit_params'][self.info['parameter']]['Upper Bound'] = cp.deepcopy(self.upper_bound_inttext.value)
            # Retrieve the fit module
            fit_module = self.info['fit_dict'][self.info['fit_type']]
            # Recalculate the parameters
            fit_module.parameters = fit_module.get_params(self.info['fit_params'])

        self.upper_bound_inttext.observe(functools.partial(upper_bound_inttext_clicked, self=self))

        #####################

        # Add a button widget to fit to the points
        self.fit_button = widgets.ToggleButton(description='Fit')

        def fit_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.fit_button: button.value = False

            # Turn on fit mode
            if self.info['interactive_mode'] == 'fit':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'fit'

        self.fit_button.observe(functools.partial(fit_button_clicked, self=self))

        #####################

        # Add a button widget to clear all points
        self.clear_axis_button = widgets.ToggleButton(description='Clear Plot')

        def clear_axis_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.clear_axis_button: button.value = False

            # Turn on clear mode
            if self.info['interactive_mode'] == 'clear':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'clear'

        self.clear_axis_button.observe(functools.partial(clear_axis_button_clicked, self=self))

        #####################

        return ([self.add_range_button,self.adjust_range_button,self.fit_type_select,self.fit_button,self.clear_axis_button],
                    [self.add_range_button,self.adjust_range_button,self.fit_button,self.clear_axis_button],
                    [self.parameter_select,self.estimate_inttext,self.lower_bound_inttext,self.upper_bound_inttext,self.fit_order_inttext])

    ##########################################
    ## UPDATE METHODS
    ##########################################

    # Method to create a new artist list
    def new_artists(self,ax,style):
        # Find the y axis minimum and maximum
        ymin,ymax = ax.get_ylim()

        artists = {
        'Bounds': ax.vlines([],ymin,ymax,linestyle=style['vline_style'],linewidth=style['vline_width']),
        'New Bound': ax.vlines([],ymin,ymax,linestyle=style['vline_style'],linewidth=style['vline_width'],color='red'),
        'Selected Bound': ax.vlines([],ymin,ymax,linestyle=style['vline_style'],linewidth=style['vline_width'],color='green')
        }
        return artists

    def data_init(self):
        return {
        'FitRange':[],
        'Fit':{'Axis':[]}
        }

    ##########################################
    ## EVENT HANDLER
    ##########################################

    def __call__(self,event):

        if not self.widget_on: return
        self.info['active_ax'] = None

        ##########################################
        ## UPDATE ARTISTS
        ##########################################

        for ax in self.fig.axes:
            # Determine which axis was clicked on
            if event.inaxes == ax: self.info['active_ax'] = ax
            # If the axis is not interactive, do not add new artists
            if ax not in self.artists_global['Interactive Axes']: continue
            # If the artist is interactive, initialize the appropriate artists
            self.update_dictionaries(ax,artist_key='Bounds',data_key='Fit')
            self.update(self.artists_global,self.data_global)

        # If a click is not within the axis, do nothing
        if self.info['active_ax'] == None: return
        else: ax = self.info['active_ax']

        ##########################################
        ## ADD RANGE
        ##########################################

        # If in add range mode
        if self.info['interactive_mode'] == 'add_range':

            # Check to see if the axis is an interactive axis
            if ax not in self.artists_global['Interactive Axes']: return
            # Check to see how many bounds are already in the axis
            if len(self.data[ax]['FitRange']) == 2: return

            # Clear all other points
            self.clear(self.artists[ax]['New Bound'],ax=ax,vlines=True)
            self.clear(self.artists[ax]['Selected Bound'],ax=ax,vlines=True)
            self.info['selected'] = False

            # Add the point to the array of boundaries
            self.data[ax]['FitRange'].append(event.xdata)

            # Replot all the points, including the extra one
            self.set_segments(self.artists[ax]['Bounds'],self.data[ax]['FitRange'],ax)

        ##########################################
        ## ADJUST RANGE
        ##########################################

        elif self.info['interactive_mode'] == 'adjust_range':

            # Check to see if the axis is an interactive axis
            if ax not in self.artists_global['Interactive Axes']: return

            # If a boundary is not selected
            if not self.info['selected']:

                # If there was a new boundary, clear it
                self.clear(self.artists[ax]['New Bound'],ax=ax,vlines=True)
                # Find the distance between the click and each boundary
                self.dat = event.xdata
                hdists = self.hdist(event.xdata,self.data[ax]['FitRange'],ax)
                self.hdists = hdists
                # Find the index of the point closest to the click
                self.info['close_bound'] = np.nanargmin(hdists)

                # If the point is close to the click
                if hdists[self.info['close_bound']] < self.info['click_dist']:

                    # Replot the selected boundary in a different color
                    self.set_segments(self.artists[ax]['Selected Bound'],[self.data[ax]['FitRange'][self.info['close_bound']]],ax)
                    # State that a point has been selected
                    self.info['selected'] = True

            # If a point has already been selected
            else:
                # Remove the bound from the data array
                self.data[ax]['FitRange'] = np.delete(self.data[ax]['FitRange'],self.info['close_bound'])
                # Remove the temporary plotted point
                self.clear(self.artists[ax]['Selected Bound'],ax=ax,vlines=True)

                # Plot the new line
                self.set_segments(self.artists[ax]['New Bound'],[event.xdata],ax)
                # Add the point to the array of points
                self.data[ax]['FitRange'] = np.append(self.data[ax]['FitRange'],event.xdata)

                # Replot all points
                self.set_segments(self.artists[ax]['Bounds'],self.data[ax]['FitRange'],ax)

                # No point is currently selected
                self.info['selected'] = False

        ##########################################
        ## MODEL FIT
        ##########################################

        # If in fit mode
        elif self.info['interactive_mode'] == 'fit':

            # Check to see if the axis is an interactive axis
            if ax not in self.artists_global['Interactive Axes']: return

            # Only fit if there are two boundaries
            if len(self.data[ax]['FitRange']) < 2: return

            if self.info['fit_type'] is not None:
                ## Import functions from a module
                imported_fit = self.info['fit_dict'][self.info['fit_type']]
                self.fit(imported_fit,ax,self.info,new_ax=imported_fit.new_ax)

        ##########################################
        ## CLEAR AXIS
        ##########################################

        # If in clear mode
        elif self.info['interactive_mode'] == 'clear':

            # If the artist is not meant for data display
            if ax in self.artists_global['Interactive Axes']:
                # If there is a fit axis
                if len(self.data[ax]['Fit']['Axis']) != 0:
                    # For each fit axis
                    for fit_ax in self.data[ax]['Fit']['Axis']:
                        # Clear data
                        del self.data[fit_ax]
                        # Clear global data
                        if fit_ax in self.data_global.keys():
                            del self.data_global[fit_ax]
                        # Clear artists
                        for artist in self.artists[fit_ax]:
                            self.artists[fit_ax][artist].set_paths([])
                        # Clear the artist dictionary entry
                        del self.artists[fit_ax]
                        # Clear global artists
                        if fit_ax in self.artists_global.keys():
                            del self.artists_global[fit_ax]
                        # Clear the primary artists
                        del self.artists_global['Primary Artists'][fit_ax]

                        # Delete axis
                        self.fig.delaxes(fit_ax)
                        self.resize_figure(len(self.fig.axes),self.fig)

                # Clear the data dictionary
                self.data[ax] = self.data_init()
                # Clear artists
                self.set_segments(self.artists[ax]['Bounds'],[],ax)
                self.set_segments(self.artists[ax]['New Bound'],[],ax)
                self.set_segments(self.artists[ax]['Selected Bound'],[],ax)

                for artist in list(self.artists[ax].keys()):
                    if 'Fit Line' in artist:
                        self.artists[ax][artist].set_data([],[])
                        del self.artists[ax][artist]

            else:
                # Clear global artists
                if ax in self.artists_global.keys():
                    del self.artists_global[ax]
                # Clear artists on parent plot
                for artist in list(self.artists[self.data[ax]['Data Axis']].keys()):
                    if artist == f"Fit Line {self.data[ax]['Fit Number']}":
                        # Clear line data
                        self.artists[self.data[ax]['Data Axis']][artist].set_data([],[])
                        del self.artists[self.data[ax]['Data Axis']][artist]
                        if self.data[ax]['Data Axis'] in self.artists_global.keys():
                            # if artist in self.artists_global[self.data[ax]['Data Axis']].keys():
                            self.artists_global[self.data[ax]['Data Axis']].pop(artist,None)

                # Clear artists
                del self.artists[ax]
                # Clear the data on the parent plot
                self.data[self.data[ax]['Data Axis']]['Fit'] = {'Line 1':{'xdata':[],'ydata':[]},'Axis':[]}
                # Clear the data on the parent plot
                if self.data[ax]['Data Axis'] in self.data_global.keys():
                    self.data[self.data[ax]['Data Axis']]['Fit'] = {'Line 1':{'xdata':[],'ydata':[]},'Axis':[]}
                # Clear data
                del self.data[ax]
                # Clear global data
                if ax in self.data_global.keys():
                    del self.data_global[ax]
                # Clear the primary artists
                del self.artists_global['Primary Artists'][ax]
                # Delete axis
                self.fig.delaxes(ax)
                self.resize_figure(len(self.fig.axes),self.fig)

            # Clear the legend
            ax.get_legend().remove()

        ##########################################
        ## UPDATE ARTISTS
        ##########################################

        # Update artists
        if self.show: plt.show()

    ##########################################
    ## ACTIVATE AND DEACTIVATE
    ##########################################

    # Overwrite the deactivation method called when the widget is turned off
    def activate(self):
        # Set the widget to the off setting
        self.widget_on = True
        # Display the widget menu
        self.place_menu(self.menu,self.button_list)
        # # Display the paramter menu
        # self.place_menu(self.param_menu,self.param_button_list)
        # Return the widget toggle
        return self.widget_on

    # Overwrite the deactivation method called when the widget is turned off
    def deactivate(self,main=False):
        # Deactivate the widget
        self.widget_on = False
        # Clear the widget menu
        self.place_menu(self.menu,[])
        # Clear the paramter menu
        self.place_menu(self.param_menu,[])
        # Return the widget toggle
        return self.widget_on

    ##########################################
    ## LOAD METHODS
    ##########################################

    def load(self,load_data):

        self.widget_on = True
        self.show = False

        class sim_event:
            def __init__(self,xdata,ydata,axis):
                self.xdata = xdata
                self.ydata = ydata
                self.inaxes = axis

        for axis in load_data['data']:
        # Here axis will take on values of 'Axis 1', 'Axis 2', etc.

            # Determine the axis on the new figure
            ax = self.fig.axes[int(axis.split(' ')[-1])-1]

            # Store the points which define the baseline
            if 'FitRange' in load_data['data'][axis]:
                range = load_data['data'][axis]['FitRange']
                self.info['interactive_mode'] = 'add_range'

                # Call the __call__ method to add the points with simulated events
                [self(sim_event(xdata,None,ax)) for xdata in range]

            else: continue

            # If there was a fit
            if 'Fit' in load_data['data'][axis]:

                fit = load_data['data'][axis]['Fit']

                # Loop through all the fits
                fit_num = 1
                while True:
                    if f'Fit {fit_num}' not in fit: break
                    else:
                        self.info['fit_type'] = fit[f'Fit {fit_num}']['fit_type']
                        self.info['fit_params'] = fit[f'Fit {fit_num}']['fit_params']
                        self.data[ax]['FitRange'] = fit[f'Fit {fit_num}']['FitRange']

                        # Apply fit order
                        self.info['fit_order'] = fit[f'Fit {fit_num}']['fit_order']
                        # Retrieve the fit module
                        fit_module = self.info['fit_dict'][self.info['fit_type']]
                        # Set the fit order in the module
                        fit_module.fit_order = self.info['fit_order']
                        # Recalculate the parameters
                        fit_module.parameters = fit_module.get_params(self.info['fit_params'])

                        self.info['interactive_mode'] = 'fit'

                        # Call the __call__ method to add the points with simulated events
                        self(sim_event(None,None,ax))

                        fit_num += 1

        self.widget_on = False
        self.show = True
        self.info['interactive_mode'] = 'off'



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 27 September 2022

from .main_menu import main_menu

class puma(main_menu):
    pass



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 22 August 2022

import numpy as np
import matplotlib.pyplot as plt
import ipywidgets as widgets
import functools

from scipy.interpolate import interp1d

from .base import widget_base

class baseline(widget_base):
    '''
    '''
    def __init__(self,fig,menu=None,artists_global=None,data_global=None,load_data=None):
        '''
        '''
        # Initialize the widget in the off position
        self.widget_on = False
        # Set the display setting to display output
        self.show = True
        # Setup the figure and axes
        self.fig = fig
        # Setup possible load data
        if load_data is not None:
            if 'baseline' in load_data: load_data = load_data['baseline']

        # Initialize all global matplotlib artists
        self.artists_global = self.pull_artists(artists_global)
        # Initialize all global data
        self.data_global = self.pull_data(data_global)
        # Create a local artist dictionary
        self.artists = {}

        # Create info and style dictionaries
        self.style,self.info,self.data = self.setup_defaults(load_data)

        # Initialize all buttons
        self.button_list,self.toggle_buttons = self.setup_buttons()
        # Place all the menu buttons
        if menu is None:
            self.menu = widgets.HBox()
            self.place_menu(self.menu,self.button_list)
        else: self.menu = menu

        if load_data is not None: self.load(load_data)

    # Define dictionaries of default values
    def setup_defaults(self,load_data=None):

        style = {
        'marker':'*',
        'markersize':15,
        'primary_color':'deepskyblue',
        'new_color':'red',
        'sel_color':'green',
        'fit_color':'darkviolet',
        'fit_color':'skyblue'}

        info = {
        # Initialize the selected boolean
        'selected':False,
        # Initialize the fit override order
        'fit_order_override':0,
        # Set the distance a click needs to be from a point to select it
        'click_dist':0.02,
        # Set a default fit type
        'fit_type':'Linear Segment',
        # Initialize the add, move, and delete mode holder
        'interactive_mode':'off',
        # Initialize an active axis
        'active_ax':None,
        # Initialize an array of axes with a baseline
        'baseline axes':[]
        }

        data = dict.fromkeys(self.artists_global['Interactive Axes'],self.data_init())

        if load_data is not None:
            load_data['info']['interactive_mode'] = 'off'
            load_data['info']['selected'] = False
            load_data['info']['active_ax'] = None

            return load_data['style'],load_data['info'],data

        return style,info,data

    # Define all buttons and their functionality
    def setup_buttons(self):

        ##########################################
        ## INTERACTIVE PLOTTING BUTTONS
        ##########################################

        # Add a button widget to add new points
        self.add_button = widgets.ToggleButton(description='Add Points')

        def on_add_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.add_button: button.value = False

            # Turn on add mode
            if self.info['interactive_mode'] == 'add':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'add'

        self.add_button.observe(functools.partial(on_add_button_clicked, self=self))

        #####################

        # Add a button widget to move points
        self.move_button = widgets.ToggleButton(description='Move Points')

        def on_move_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.move_button: button.value = False

            # Turn on move mode
            if self.info['interactive_mode'] == 'move':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'move'


        self.move_button.observe(functools.partial(on_move_button_clicked, self=self))

        #####################

        # Add a button widget to delete points
        self.delete_button = widgets.ToggleButton(description='Delete Points')

        def on_delete_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.delete_button: button.value = False

            # Turn on delete mode
            if self.info['interactive_mode'] == 'delete':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'delete'

        self.delete_button.observe(functools.partial(on_delete_button_clicked, self=self))

        ##########################################

        # Add a dropdown box to select the fit type
        self.fit_type_select = widgets.Select(description='Functional Form:',options=['Linear Segment','Polynomial'],rows=0,style={'description_width': 'initial'},layout = widgets.Layout(width='245px'))

        def on_fit_type_selected(b, self = self):

            # Change fit order value
            self.info['fit_type'] = self.fit_type_select.value
            if self.info['fit_type'] == 'Polynomial':
                self.button_list = [self.add_button,self.move_button,self.delete_button,self.fit_type_select,self.fit_order_inttext,self.clear_button]
                self.place_menu(self.menu,self.button_list)
            elif self.info['fit_type'] == 'Linear Segment':
                self.button_list = [self.add_button,self.move_button,self.delete_button,self.fit_type_select,self.clear_button]
                self.place_menu(self.menu,self.button_list)
            # Recalculate the fit
            for ax in self.info['baseline axes']:
                self.fit(ax)

        self.fit_type_select.observe(functools.partial(on_fit_type_selected, self=self))

        ##########################################

        # Add a text box to set the fit order
        self.fit_order_inttext = widgets.IntText(description='Order:',value=0,style={'description_width': 'initial'},layout = widgets.Layout(width='100px'))

        def on_fit_order_inttext_entered(b, self = self):

            # Change fit order value
            self.info['fit_order_override'] = self.fit_order_inttext.value
            # Recalculate the fit
            for ax in self.info['baseline axes']:
                self.fit(ax)

        self.fit_order_inttext.observe(functools.partial(on_fit_order_inttext_entered, self=self))

        #####################

        # Add a button widget to clear all points
        self.clear_button = widgets.ToggleButton(description='Clear Plot')

        def on_clear_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.clear_button: button.value = False

            # Turn on clear mode
            if self.info['interactive_mode'] == 'clear':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'clear'

        self.clear_button.observe(functools.partial(on_clear_button_clicked, self=self))

        #####################

        return [self.add_button,self.move_button,self.delete_button,self.fit_type_select,self.clear_button], \
        [self.add_button,self.move_button,self.delete_button,self.clear_button]

    ##########################################
    ## UPDATE METHODS
    ##########################################

    # Initialize Artists
    def new_artists(self,ax,style):

        # Plot
        bspoints, = ax.plot([],[],marker=style['marker'],linestyle='',color=style['primary_color'],markersize=style['markersize'])
        newpoint, = ax.plot([],[],color=style['new_color'],marker=style['marker'],markersize=style['markersize'])
        selpoint, = ax.plot([],[],color=style['sel_color'],marker=style['marker'],markersize=style['markersize'])
        plotfit, = ax.plot([],[],color=style['fit_color'])

        artists = {
        'Baseline Points': bspoints,
        'New Points': newpoint,
        'Selected Points': selpoint,
        'Baseline Func': plotfit
        }
        return artists

    def data_init(self):
        return {
            'Baseline':{
                # Fit points
                'Baseline Points': {'xdata':[],'ydata':[]},
                'Baseline': {'xdata':[],'ydata':[]},
                'Corrected Data': {'xdata':[],'ydata':[]},
                'Baseline Axis':None
                }
        }

    ##########################################
    ## EVENT HANDLER
    ##########################################

    # When the class is called as a function
    def __call__(self,event):

        if not self.widget_on: return
        self.info['active_ax'] = None

        ##########################################
        ## UPDATE ARTISTS
        ##########################################

        for ax in self.fig.axes:
            # Determine which axis was clicked on
            if event.inaxes == ax: self.info['active_ax'] = ax
            # If the axis is not interactive, do not add new artists
            if ax not in self.artists_global['Interactive Axes']: continue
            # If the artist is interactive, initialize the appropriate artists
            self.update_dictionaries(ax,artist_key='Baseline Points',data_key='Baseline')
            self.update(self.artists_global,self.data_global)

        # If a click is not within the axis, do nothing
        if self.info['active_ax'] == None: return
        else: ax = self.info['active_ax']

        ##########################################
        ## ADD POINTS
        ##########################################

        # If the add button has been pressed
        if self.info['interactive_mode'] == 'add':

            # Clear all other points
            self.clear(self.artists[ax]['New Points'])
            self.clear(self.artists[ax]['Selected Points'])
            self.info['selected'] = False

            # Add the point to the array of points
            self.data[ax]['Baseline']['Baseline Points']['xdata'] = np.append(self.data[ax]['Baseline']['Baseline Points']['xdata'],event.xdata)
            self.data[ax]['Baseline']['Baseline Points']['ydata'] = np.append(self.data[ax]['Baseline']['Baseline Points']['ydata'],event.ydata)

            # Replot all the points, including the extra one
            self.artists[ax]['Baseline Points'].set_data([self.data[ax]['Baseline']['Baseline Points']['xdata'],self.data[ax]['Baseline']['Baseline Points']['ydata']])

        ##########################################
        ## MOVE POINTS
        ##########################################

        # If the move button has been pressed
        elif self.info['interactive_mode'] == 'move':

            # If a point is not selected
            if not self.info['selected']:
                # If there was a newpoint added, clear it
                self.clear(self.artists[ax]['New Points'])
                # Calculate the distance from the click to each point
                euc = self.bigeuc(event.xdata,event.ydata,self.data[ax]['Baseline']['Baseline Points']['xdata'],self.data[ax]['Baseline']['Baseline Points']['ydata'],ax)
                # Find the index of the point closest to the click
                self.close = np.nanargmin(euc)

                # If the point is close to the click
                if euc[self.close] < self.info['click_dist']:

                    # Replot the selected point in a different color
                    self.artists[ax]['Selected Points'].set_data([self.data[ax]['Baseline']['Baseline Points']['xdata'][self.close],self.data[ax]['Baseline']['Baseline Points']['ydata'][self.close]])
                    # State that a point has been selected
                    self.info['selected'] = True

            # If a point has already been selected
            else:
                # Remove the point from the data arrays
                self.data[ax]['Baseline']['Baseline Points']['xdata'] = np.delete(self.data[ax]['Baseline']['Baseline Points']['xdata'],self.close)
                self.data[ax]['Baseline']['Baseline Points']['ydata'] = np.delete(self.data[ax]['Baseline']['Baseline Points']['ydata'],self.close)

                # Remove the temporary plotted point
                self.clear(self.artists[ax]['Selected Points'])

                # Plot the newly placed point
                self.artists[ax]['New Points'].set_data([event.xdata, event.ydata])

                # Add the point to the array of points
                self.data[ax]['Baseline']['Baseline Points']['xdata'] = np.append(self.data[ax]['Baseline']['Baseline Points']['xdata'],event.xdata)
                self.data[ax]['Baseline']['Baseline Points']['ydata'] = np.append(self.data[ax]['Baseline']['Baseline Points']['ydata'],event.ydata)

                # Replot all the points, excluding the selected one
                self.artists[ax]['Baseline Points'].set_data([self.data[ax]['Baseline']['Baseline Points']['xdata'],self.data[ax]['Baseline']['Baseline Points']['ydata']])

                # No point is currently selected
                self.info['selected'] = False

        ##########################################
        ## DELETE POINTS
        ##########################################

        # If the delete button has been pressed
        elif self.info['interactive_mode'] == 'delete':

            # Clear all other points
            self.clear(self.artists[ax]['New Points'])
            self.clear(self.artists[ax]['Selected Points'])
            self.info['selected'] = False

            # Calculate the distance from the click to each point
            euc = self.bigeuc(event.xdata,event.ydata,self.data[ax]['Baseline']['Baseline Points']['xdata'],self.data[ax]['Baseline']['Baseline Points']['ydata'],ax)
            # Find the index of the point closest to the click
            close = np.nanargmin(euc)

            # If the point is close to the click
            if euc[close] < self.info['click_dist']:
                # Remove the point from the data arrays
                self.data[ax]['Baseline']['Baseline Points']['xdata'] = np.delete(self.data[ax]['Baseline']['Baseline Points']['xdata'],close)
                self.data[ax]['Baseline']['Baseline Points']['ydata'] = np.delete(self.data[ax]['Baseline']['Baseline Points']['ydata'],close)

                # Replot all the points, excluding the selected one
                self.artists[ax]['Baseline Points'].set_data([self.data[ax]['Baseline']['Baseline Points']['xdata'],self.data[ax]['Baseline']['Baseline Points']['ydata']])

            # If there is only one point left, clear the fit
            if len(self.data[ax]['Baseline']['Baseline Points']['xdata']) < 2: self.clearfit()

        ##########################################
        ## CLEAR AXIS
        ##########################################

        # If the clear button has been pressed
        elif self.info['interactive_mode'] == 'clear':
            self.axis_clear(ax)

        ##########################################
        ## UPDATE ARTISTS
        ##########################################

        # If there is a fit, update it
        self.fit(ax)
        if self.show: plt.show()

    ##########################################
    ## FITTING METHODS
    ##########################################

    # Global fit function
    def fit(self,ax):
        # If there is only one point, don't fit
        if len(self.data[ax]['Baseline']['Baseline Points']['xdata']) < 2: return
        # Fit the baseline points depending using the specified function
        if self.info['fit_type'] == 'Polynomial': self.pfit(ax)
        elif self.info['fit_type'] == 'Linear Segment': self.lsfit(ax)
        # Setup baseline if not already setup
        if self.data[ax]['Baseline']['Baseline Axis'] is None:
            # Add axis to a list of axes with baselines
            if ax not in self.info['baseline axes']: self.info['baseline axes'].append(ax)
            # Add an axis to the figure
            self.resize_figure(len(self.fig.axes)+1,self.fig)
            # Save that axis
            baseline_ax = self.fig.axes[-1]
            self.data[ax]['Baseline']['Baseline Axis'] = baseline_ax
            # Add it to the interactive figure list
            self.artists_global['Interactive Axes'].append(baseline_ax)
            # Add a baselined plot
            self.artists[baseline_ax] = {}
            self.artists[baseline_ax]['Baselined Data'] = baseline_ax.plot([],[],color=self.style['primary_color'])[0]
            self.artists_global['Primary Artists'][baseline_ax] = 'Baselined Data'
        # # Update the baseline
        self.baseline_plot(ax)

    # Polynomial fit function
    def pfit(self,ax):
        # Set the fit order to one below the number of points
        self.fit_order = np.min([5,len(self.data[ax]['Baseline']['Baseline Points']['xdata'])-1])
        # If a fit order override is specified, override the fit order
        if self.info['fit_order_override'] != 0: self.fit_order = self.info['fit_order_override']
        # Calculate a polynomial fit
        z = np.polyfit(self.data[ax]['Baseline']['Baseline Points']['xdata'],self.data[ax]['Baseline']['Baseline Points']['ydata'], self.fit_order)
        self.fitfunc = np.poly1d(z)

        # Calculate the fit arrays
        xnew = np.linspace(ax.get_xlim()[0],ax.get_xlim()[1],200)
        ynew = self.fitfunc(xnew)

        # Save the baseline
        self.data[ax]['Baseline']['Baseline']['xdata'] = xnew
        self.data[ax]['Baseline']['Baseline']['ydata'] = ynew
        # Plot the fit
        self.artists[ax]['Baseline Func'].set_data([xnew,ynew])

        # Make sure the fit degree is visually up to date
        self.fit_order_inttext.value = self.fit_order

    # Linear segmented fit function
    def lsfit(self,ax):
        # Calculate ordered copies of xdat and ydat
        xord,yord = self.order(self.data[ax]['Baseline']['Baseline Points']['xdata'],self.data[ax]['Baseline']['Baseline Points']['ydata'])
        # Create the fitting function
        self.fitfunc = interp1d(xord,yord)

        # Get the length of the data array
        bsx,_ = self.artists_global[ax][self.artists_global['Primary Artists'][ax]].get_data()
        dat_length = len(np.squeeze(np.where((bsx > xord[0]) & (bsx < xord[-1]))))
        # Create x and y data using the fitting function
        xnew = np.linspace(np.min(xord),np.max(xord),dat_length)
        ynew = self.fitfunc(xnew)

        # Save the baseline
        self.data[ax]['Baseline']['Baseline']['xdata'] = xnew
        self.data[ax]['Baseline']['Baseline']['ydata'] = ynew
        # Plot the fit
        self.artists[ax]['Baseline Func'].set_data([xnew,ynew])

    # Baseline method
    def baseline_plot(self,ax):
        # Subset the x range to what is visible
        xord,yord = self.order(self.data[ax]['Baseline']['Baseline Points']['xdata'],self.data[ax]['Baseline']['Baseline Points']['ydata'])
        # Subset the science data with the range defined by the fit
        bsx,bsy = self.artists_global[ax][self.artists_global['Primary Artists'][ax]].get_data()
        ii = np.squeeze(np.where((bsx > xord[0]) & (bsx < xord[-1])))
        # Redefine the baseline vars as their relevant subsets
        bsx,bsy = bsx[ii],bsy[ii]
        # Calculate the fit using the baseline x data
        bs_fity = self.fitfunc(bsx)
        # Save the baseline-subtracted data
        self.data[ax]['Baseline']['Corrected Data']['xdata'] = bsx
        self.data[ax]['Baseline']['Corrected Data']['ydata'] = bsy - bs_fity
        # Show the residual plot
        self.artists[self.data[ax]['Baseline']['Baseline Axis']]['Baselined Data'].set_data([bsx, bsy - bs_fity])
        # Autoscale the y axis correspondingly
        self.data[ax]['Baseline']['Baseline Axis'].relim()
        self.data[ax]['Baseline']['Baseline Axis'].autoscale()

    def axis_clear(self,ax):
        # If there is a baseline axis that was generated from this one
        if self.data[ax]['Baseline']['Baseline Axis'] is not None:
            # Recursively clear it
            self.axis_clear(self.data[ax]['Baseline']['Baseline Axis'])

            # Clear baseline axis entry in artists dict
            del self.artists[self.data[ax]['Baseline']['Baseline Axis']]
            # Clear baseline axis entry in artists dict
            del self.data[self.data[ax]['Baseline']['Baseline Axis']]
            # Clear baseline axis entry in global artists dict
            if self.data[ax]['Baseline']['Baseline Axis'] in self.artists_global.keys():
                del self.artists_global[self.data[ax]['Baseline']['Baseline Axis']]
                self.artists_global['Interactive Axes'].remove(self.data[ax]['Baseline']['Baseline Axis'])
            # Clear baseline axis entry in global artists dict
            if self.data[ax]['Baseline']['Baseline Axis'] in self.data_global.keys():
                del self.data_global[self.data[ax]['Baseline']['Baseline Axis']]

            # Remove the baseline axis
            self.fig.delaxes(self.data[ax]['Baseline']['Baseline Axis'])
            del self.artists_global['Primary Artists'][self.data[ax]['Baseline']['Baseline Axis']]
            self.data[ax]['Baseline']['Baseline Axis'] = None
            self.resize_figure(len(self.fig.axes),self.fig)

        # Clear specified artists
        for artist in self.artists[ax]:
            if artist != 'Baselined Data':
                self.clear(self.artists[ax][artist])

        # Reset selected boolean
        self.info['selected'] = False

        # Clear data dictionary
        self.data[ax] = self.data_init()

    ##########################################
    ## LOAD METHODS
    ##########################################

    def load(self,load_data):

        self.widget_on = True
        self.show = False
        self.info['interactive_mode'] = 'add'

        class sim_event:
            def __init__(self,xdata,ydata,axis):
                self.xdata = xdata
                self.ydata = ydata
                self.inaxes = axis

        for axnum,axis in enumerate(load_data['data']):
        # Here axis will take on values of 'Axis 1', 'Axis 2', etc.

            # If new axes have yet to be added, return
            if axnum >= len(self.fig.axes): break

            # Determine the axis on the new figure
            ax = self.fig.axes[int(axis.split(' ')[-1])-1]

            # Store the points which define the baseline
            if 'Baseline' in load_data['data'][axis]:
                baseline_pts = load_data['data'][axis]['Baseline']['Baseline Points']

                # Call the baseline __call__ method to add the points with simulated events
                [self(sim_event(xdata,ydata,ax)) for xdata,ydata in zip(baseline_pts['xdata'],baseline_pts['ydata'])]

        self.widget_on = False
        self.show = True
        self.info['interactive_mode'] = 'off'



# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 23 September 2022

import numpy as np
import matplotlib.pyplot as plt
import ipywidgets as widgets
import functools

import seaborn as sns

from .base import widget_base

class labels(widget_base):
    def __init__(self,fig,menu=None,data=None,artists_global=None,data_global=None,load_data=None):
        '''
        '''
        # Initialize the widget in the off position
        self.widget_on = False
        # Set the display setting to display output
        self.show = True
        # Setup the figure
        self.fig = fig
        # Setup possible load data
        if load_data is not None:
            if 'labels' in load_data: load_data = load_data['labels']

        # Initialize all global matplotlib artists
        self.artists_global = self.pull_artists(artists_global)
        # Initialize all global data
        self.data_global = self.pull_data(data_global)
        # Create a dictionary using the axes as keys
        self.artists = {}

        # Initialize defaults
        self.style,self.info,self.data = self.setup_defaults(load_data)

        # Initialize all buttons
        self.button_list,self.toggle_buttons = self.setup_buttons()
        # Place all the menu buttons
        if menu is None:
            self.menu = widgets.HBox()
            self.place_menu(self.menu,self.button_list)
        else: self.menu = menu

        if load_data is not None: self.load(load_data)

    def setup_defaults(self,load_data=None):

        style = {
        'fill_color':'deepskyblue',
        'alpha':0.6,
        'vline_style':'--',
        'vline_width':2,
        'color_palette':'Set2',
        'peak_style':'-',
        'peak_width':1,
        'selsize':{'Arrows':10,'Labels':40},
        'lab_size':12
        }

        info = {
        # Initialize the add, move, and delete
        'interactive_mode':'off',
        # Initialize an active axis
        'active_ax':None,
        # Initialize a selected boolean
        'selected':False,
        # Set an interactive click distance
        'click_dist':0.04,
        # Save the number of arrows
        'arrow_num':0,
        # Save the number of labels
        'label_num':0,
        # Initialize an arrow base boolean
        'arrow_based':False,
        # Initialize label text
        'label_text':''
        }

        data = dict.fromkeys(self.artists_global['Interactive Axes'],self.data_init())

        if load_data is not None:
            load_data['info']['interactive_mode'] = 'off'
            load_data['info']['selected'] = False
            load_data['info']['active_ax'] = None

            return load_data['style'],load_data['info'],data

        return style,info,data

    def setup_buttons(self):

        ##########################################
        ## INTERACTIVE PLOTTING BUTTONS
        ##########################################

        # Add a button widget to add new points
        self.add_arrow_button = widgets.ToggleButton(description='Add Arrow')

        def on_add_arrow_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.add_arrow_button: button.value = False

            # Turn on add_range mode
            if self.info['interactive_mode'] == 'add_arrow':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'add_arrow'

        self.add_arrow_button.observe(functools.partial(on_add_arrow_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.add_label_button = widgets.ToggleButton(description='Add Label')

        def add_label_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.add_label_button: button.value = False

            # Turn on integrate mode
            if self.info['interactive_mode'] == 'add_label':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'add_label'

        self.add_label_button.observe(functools.partial(add_label_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.label_text = widgets.Text(description='Label:',style={'description_width':'initial'},layout = widgets.Layout(width='115px'))

        def label_text_clicked(b, self = self):

            if self.label_text.value != 0:
                self.info['label_text'] = self.label_text.value
            else: self.info['label_text'] = ''

        self.label_text.observe(functools.partial(label_text_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.label_size_inttext = widgets.IntText(description='Size:',value=12,style={'description_width':'initial'},layout = widgets.Layout(width='80px'))

        def label_size_inttext_clicked(b, self = self):

            if self.label_size_inttext.value != 0:
                self.style['lab_size'] = self.label_size_inttext.value
            else: self.style['lab_size'] = 12

        self.label_size_inttext.observe(functools.partial(label_size_inttext_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.adjust_button = widgets.ToggleButton(description='Adjust')

        def on_adjust_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.adjust_button: button.value = False

            # Turn on adjust_range mode
            if self.info['interactive_mode'] == 'adjust':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'adjust'

        self.adjust_button.observe(functools.partial(on_adjust_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.delete_button = widgets.ToggleButton(description='Delete')

        def delete_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.delete_button: button.value = False

            # Turn on find peaks mode
            if self.info['interactive_mode'] == 'delete':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'delete'

        self.delete_button.observe(functools.partial(delete_button_clicked, self=self))

        #####################

        # Add a button widget to add new points
        self.clear_axis_button = widgets.ToggleButton(description='Clear Plot')

        def clear_axis_button_clicked(b, self = self):

            # Turn off the other buttons
            for button in self.toggle_buttons:
                if button != self.clear_axis_button: button.value = False

            # Turn on clear mode
            if self.info['interactive_mode'] == 'clear':
                self.info['interactive_mode'] = 'off'
            else: self.info['interactive_mode'] = 'clear'

        self.clear_axis_button.observe(functools.partial(clear_axis_button_clicked, self=self))

        #####################

        return [self.add_arrow_button,self.add_label_button,self.label_text,self.label_size_inttext,self.adjust_button,self.delete_button,self.clear_axis_button],[self.add_arrow_button,self.add_label_button,self.adjust_button,self.delete_button,self.clear_axis_button]

    ##########################################
    ## UPDATE METHODS
    ##########################################

    # Method to create a new artist list
    def new_artists(self,ax,style):
        # Find the center of the axis
        xavg = np.mean(ax.get_xlim())
        yavg = np.mean(ax.get_ylim())

        # Find the arrow and label numbers
        arrow_num = self.info['arrow_num']
        label_num = self.info['label_num']

        artists = {}
        # Add artists if arrows or labels have been added
        if arrow_num != 0: artists[f'Arrow {arrow_num}'] = ax.annotate('',xy=[xavg,yavg],arrowprops=dict(arrowstyle="->")),
        if label_num != 0: artists[f'Label {label_num}'] = ax.annotate('',xy=[xavg,yavg],size=self.style['lab_size'])

        return artists

    def data_init(self):
        return {
        'Arrows':{},
        'Labels':{}
        }

    ##########################################
    ## EVENT HANDLER
    ##########################################

    def __call__(self,event):

        if not self.widget_on: return
        self.info['active_ax'] = None

        ##########################################
        ## UPDATE ARTISTS
        ##########################################

        for ax in self.fig.axes:
            # Determine which axis was clicked on
            if event.inaxes == ax: self.info['active_ax'] = ax
            # # Initialize the appropriate artists
            self.update_dictionaries(ax,artist_key='Arrow Base',data_key='Arrows')
            self.update(self.artists_global,self.data_global)
            # Add an arrow base to the artists dictionary if it does not yet exist
            if 'Arrow Base' not in self.artists[ax].keys():
                self.artists[ax]['Arrow Base'] = ax.plot([],[],marker='.',linestyle='',color='red',markersize=12)[0]
                self.artists[ax]['New Arrow/Label'] = ax.plot([],[],markeredgecolor='r',marker='o',markerfacecolor='None')[0]
                self.artists[ax]['Selected Arrow/Label'] = ax.plot([],[],markeredgecolor='g',marker='o',markerfacecolor='None')[0]

        # If a click is not within the axis, do nothing
        if self.info['active_ax'] == None: return
        else: ax = self.info['active_ax']

        # Clear all extra artists
        self.clear(self.artists[ax]['New Arrow/Label'],ax=ax)
        self.clear(self.artists[ax]['Selected Arrow/Label'],ax=ax)

        ##########################################
        ## ADD ARROW
        ##########################################

        # If in add range mode
        if self.info['interactive_mode'] == 'add_arrow':

            if self.info['arrow_based']:

                # Get the arrow number that is being added to
                arrow_num = self.info['arrow_num']
                # Save the data of the arrow head
                arrow_dat = self.data[ax]['Arrows'][f'Arrow {arrow_num}']
                arrow_dat['Head'] = {'xdata':event.xdata,'ydata':event.ydata}
                # Remove the base point artist
                self.clear(self.artists[ax]['Arrow Base'],ax=ax)
                # Flip the arrow boolean
                self.info['arrow_based'] = False

                # Add an arrow annotation
                basex = self.data[ax]['Arrows'][f'Arrow {arrow_num}']['Base']['xdata']
                basey = self.data[ax]['Arrows'][f'Arrow {arrow_num}']['Base']['ydata']

                # Set the head of the arrow
                self.artists[ax][f'Arrow {arrow_num}'] = ax.annotate('',xy=(event.xdata,event.ydata),xytext=(basex,basey),arrowprops=dict(arrowstyle="->"))

            # If this is a new arrow
            else:
                # Increase the number of complete arrows
                self.info['arrow_num'] += 1
                arrow_num = self.info['arrow_num']
                # Flip the arrow boolean
                self.info['arrow_based'] = True
                # Save the location of the arrow base
                self.data[ax]['Arrows'][f'Arrow {arrow_num}'] = {'Base':{'xdata':event.xdata,'ydata':event.ydata}}
                # Plot a point where the base is
                self.artists[ax]['Arrow Base'].set_data(event.xdata,event.ydata)

        ##########################################
        ## ADD LABEL
        ##########################################

        # If in add label mode
        if self.info['interactive_mode'] == 'add_label':

            # If the label is going to be empty, dont bother
            if self.info['label_text'] == '': return
            # Increment the number of labels
            self.info['label_num'] += 1
            # Get the arrow number that is being added to
            label_num = self.info['label_num']
            # Return the label offsets
            xoffset,yoffset = self.get_label_centers(self.info['label_text'],event,ax)
            # Replot the label with the offsets applied
            self.artists[ax][f'Label {label_num}'] = ax.annotate(self.info['label_text'],xy=(event.xdata+xoffset,event.ydata+yoffset/2),size=self.style['lab_size'])
            # Save label location and text data
            self.data[ax]['Labels'][f'Label {label_num}'] = {'xdata':event.xdata,'ydata':event.ydata-yoffset/2,'label_text':self.info['label_text']}

        ##########################################
        ## ADJUST ARROWS/LABELS
        ##########################################

        elif self.info['interactive_mode'] == 'adjust':

            # If an artist is not selected
            if not self.info['selected']:

                # If there was a half added arrow, clear the data
                self.clear(self.artists[ax]['Arrow Base'],ax=ax)
                # Flip the arrow boolean
                self.info['arrow_based'] = False

                # Get all the arrow and label locations
                xlocs,ylocs,labels = self.get_dists(ax)

                # Calculate the distance from the click to each point
                euc = self.bigeuc(event.xdata,event.ydata,xlocs,ylocs,ax)
                # Find the index of the point closest to the click
                self.info['close_index'] = np.nanargmin(euc)

                # If the point is close to the click
                if euc[self.info['close_index']] < self.info['click_dist']:

                    # Plot a circle on top of the selected arrow
                    if 'Arrow' in labels[self.info['close_index']]:
                        self.artists[ax]['Selected Arrow/Label'].set_data(xlocs[self.info['close_index']],ylocs[self.info['close_index']])
                        self.artists[ax]['Selected Arrow/Label'].set(markersize=self.style['selsize']['Arrows'])

                    # Or plot on top of the selected label
                    if 'Label' in labels[self.info['close_index']]:

                        # Calculate the center offsets for the label
                        xoffset,yoffset = self.get_label_centers(self.data[ax]['Labels'][labels[self.info['close_index']]]['label_text'],event,ax)

                        # Plot the selected circle
                        self.artists[ax]['Selected Arrow/Label'].set_data(xlocs[self.info['close_index']],ylocs[self.info['close_index']]+yoffset/4)
                        self.artists[ax]['Selected Arrow/Label'].set(markersize=self.style['selsize']['Labels'])

                    # State that an artist has been selected
                    self.info['selected'] = True

            # If an artist has been selected
            else:
                # Get all the arrow and label locations
                xlocs,ylocs,labels = self.get_dists(ax)

                if 'Arrow' in labels[self.info['close_index']]:
                    # Set a label type
                    lab_type = 'Arrows'
                    # Find the name and base/head value of the arrow
                    arr_name,end = labels[self.info['close_index']].split(':')
                    # Find the data dictionary corresponding to the selected arrow
                    labdata = self.data[ax][lab_type][arr_name][end[1:]]
                    # Set the new location data
                    labdata['xdata'],labdata['ydata'] = event.xdata,event.ydata

                    # Get the data for the arrow
                    if end == ' Head':
                        hxdata,hydata = event.xdata,event.ydata
                        bxdata,bydata = self.data[ax][lab_type][arr_name]['Base']['xdata'],self.data[ax][lab_type][arr_name]['Base']['ydata']
                    elif end == ' Base':
                        hxdata,hydata = self.data[ax][lab_type][arr_name]['Head']['xdata'],self.data[ax][lab_type][arr_name]['Head']['ydata']
                        bxdata,bydata = event.xdata,event.ydata

                    # Delete the previous arrow artist
                    self.artists[ax][arr_name].remove()
                    # Replot the arrow
                    self.artists[ax][arr_name] = ax.annotate('',xy=(hxdata,hydata),xytext=(bxdata,bydata),arrowprops=dict(arrowstyle="->"))

                    # Set the new artist offset to 0
                    yoffset = 0

                else:
                    # Set a label type
                    lab_type = 'Labels'
                    # Find the name of the label
                    lab_name = labels[self.info['close_index']]
                    # Find the data dictionary corresponding to the selected label
                    labdata = self.data[ax][lab_type][lab_name]

                    # Calculate the center offsets for the label
                    xoffset,yoffset = self.get_label_centers(labdata['label_text'],event,ax)

                    # Set the new location data
                    labdata['xdata'],labdata['ydata'] = event.xdata,event.ydata-yoffset/2

                    # Delete the previous label artist
                    self.artists[ax][lab_name].remove()

                    # Replot the label
                    self.artists[ax][lab_name] = ax.annotate(labdata['label_text'],xy=(event.xdata+xoffset,event.ydata+yoffset/2),size=self.style['lab_size'])

                # Plot a red circle around the new arrow
                self.artists[ax]['New Arrow/Label'].set_data(event.xdata,event.ydata-yoffset/4)
                self.artists[ax]['New Arrow/Label'].set(markersize=self.style['selsize'][lab_type])

                # No point is currently selected
                self.info['selected'] = False

        ##########################################
        ## DELETE POINTS
        ##########################################

        # If the delete button has been pressed
        elif self.info['interactive_mode'] == 'delete':

            # Set the selected boolean to false
            self.info['selected'] = False

            # Get all the arrow and label locations
            xlocs,ylocs,labels = self.get_dists(ax)

            # Calculate the distance from the click to each point
            euc = self.bigeuc(event.xdata,event.ydata,xlocs,ylocs,ax)
            # Find the index of the point closest to the click
            self.info['close_index'] = np.nanargmin(euc)

            # If the point is close to the click
            if euc[self.info['close_index']] < self.info['click_dist']:

                # Set a label type
                if 'Arrow' in labels[self.info['close_index']]:
                    lab_type = 'Arrows'
                    lab_name = labels[self.info['close_index']].split(':')[0]
                else:
                    lab_type = 'Labels'
                    lab_name = labels[self.info['close_index']]

                # Remove the label from the local data array
                del self.data[ax][lab_type][lab_name]
                # Remove the label from the global data array
                if lab_name in self.data_global[ax][lab_type].keys():
                    del self.data_global[ax][lab_type][lab_name]

                # Remove the artist from the plot
                self.artists[ax][lab_name].remove()
                # Remove the label from the local artist array
                del self.artists[ax][lab_name]
                # Remove the label from the global artist array
                if lab_name in self.artists_global[ax].keys():
                    del self.artists_global[ax][lab_name]


        ##########################################
        ## CLEAR AXIS
        ##########################################

        # If in clear mode
        elif self.info['interactive_mode'] == 'clear':

            for lab_name in list(self.artists[ax].keys()):
                # Remove the artist from the plot
                if 'Arrow ' in lab_name or 'Label ' in lab_name:
                    self.artists[ax][lab_name].remove()
                    # Remove the label from the local artist array
                    del self.artists[ax][lab_name]
                    # Remove the label from the global artist array
                    if lab_name in self.artists_global[ax].keys():
                        del self.artists_global[ax][lab_name]

            for lab_type in ['Arrows','Labels']:
                # Remove the label from the local data array
                del self.data[ax][lab_type]
                # Remove the label from the global data array
                if lab_type in self.data_global[ax].keys():
                    del self.data_global[ax][lab_type]

            # Set the selected boolean to false
            self.info['selected'] = False
            # Reset the arrow number
            self.info['arrow_num'] = 0
            # Reset the label number
            self.info['label_num'] = 0

        ##########################################
        ## UPDATE ARTISTS
        ##########################################

        # Update artists
        if self.show: plt.show()

    def get_dists(self,ax):
        # Define a new dictionary to hold the distances
        dist_dict = {}
        # For Arrows and Labels in the data dictionary
        for arr_lab in self.data[ax].keys():
            # For each arrow or each label
            for key in self.data[ax][arr_lab]:
                # If it is an arrow
                if 'Arrow ' in key:
                    # Save the arrow data to the distance dict
                    dist_dict[f'{key}: Base'] = [self.data[ax][arr_lab][key]['Base']['xdata'],self.data[ax][arr_lab][key]['Base']['ydata']]
                    dist_dict[f'{key}: Head'] = [self.data[ax][arr_lab][key]['Head']['xdata'],self.data[ax][arr_lab][key]['Head']['ydata']]
                # If it is a label
                if 'Label ' in key:
                    # Save the label data to the distance dict
                    dist_dict[key] = [self.data[ax][arr_lab][key]['xdata'],self.data[ax][arr_lab][key]['ydata']]
        # Save the labels for each distance
        labs = list(dist_dict.keys())
        # Zip the x,y tuples into x and y arrays
        x,y = zip(*dist_dict.values())
        # Return the x, y, and label array
        return x,y,labs

    def get_label_offsets(self,label_artist,ax):
        # Get the bounding box bounds for the label artist
        bb = np.array(label_artist.get_tightbbox(None).bounds)
        # Caclulate the figure pixel coords of the lower left and upper right corners
        low_left = bb[:2]; up_right = bb[:2] + bb[2:]
        # Define a figure pixel inversion transformation
        inv = ax.transData.inverted()
        # Average the corners to find the x and y offsets of the box in data coordinates
        xoffset,yoffset = (inv.transform(low_left) - inv.transform(up_right))/2
        # Return the label offsets
        return xoffset,yoffset

    def get_label_centers(self,text,event,ax):
        # Set a dummy label
        dummy_label = ax.annotate(text,xy=(event.xdata,event.ydata),size=self.style['lab_size'],color='None')
        # Calculate offsets from the lower left corner to the label center
        xoffset,yoffset = self.get_label_offsets(dummy_label,ax=ax)
        # Remove the dummy artist from the plot
        dummy_label.remove()
        # Return the label offsets
        return xoffset,yoffset

    ##########################################
    ## LOAD METHODS
    ##########################################

    def load(self,load_data):

        self.widget_on = True
        self.show = False

        class sim_event:
            def __init__(self,xdata,ydata,axis):
                self.xdata = xdata
                self.ydata = ydata
                self.inaxes = axis

        for axnum,axis in enumerate(load_data['data']):
        # Here axis will take on values of 'Axis 1', 'Axis 2', etc.

            # If new axes have yet to be added, return
            if axnum >= len(self.fig.axes): break

            # Determine the axis on the new figure
            ax = self.fig.axes[int(axis.split(' ')[-1])-1]

            # Store the points which define the baseline
            if 'Arrows' in load_data['data'][axis]:

                # Shorten the path
                arr_dat = load_data['data'][axis]['Arrows']

                # Loop through all the arrows
                arrow_num = 1
                while True:
                    if f'Arrow {arrow_num}' not in arr_dat: break
                    else:
                        adata = arr_dat[f'Arrow {arrow_num}']

                        self.info['interactive_mode'] = 'add_arrow'

                        # Call the __call__ method to add the tail point
                        self(sim_event(adata['Base']['xdata'],adata['Base']['ydata'],ax))

                        # Call the __call__ method to add the head point
                        self(sim_event(adata['Head']['xdata'],adata['Head']['ydata'],ax))

                        arrow_num += 1

            # Store the points which define the baseline
            if 'Labels' in load_data['data'][axis]:

                # Shorten the path
                lab_dat = load_data['data'][axis]['Labels']

                # Loop through all the arrows
                lab_num = 1
                while True:
                    if f'Label {lab_num}' not in lab_dat: break
                    else:
                        ldata = lab_dat[f'Label {lab_num}']

                        self.info['interactive_mode'] = 'add_label'
                        self.info['label_text'] = ldata['label_text']

                        # Call the __call__ method to add the tail point
                        self(sim_event(ldata['xdata'],ldata['ydata'],ax))

                        lab_num += 1

        self.widget_on = False
        self.show = True
        self.info['interactive_mode'] = 'off'



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="SpacePuma",
    version="0.21",
    author="Alex DelFranco",
    author_email="adelfranco24@amherst.edu",
    description=("Interact with your data."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alexdelfranco/SpacePuma",
    packages=["spacepuma"],
    license="MIT",
    install_requires=[
        "numpy",
        "scipy",
        "jupyterlab",
        "matplotlib",
        "seaborn",
        "ipympl",
        ],
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
